#include "../common/pkmem-user.h"
#include "../common/registry.h"
#include "pc-pkfifo.h"

int main(int argc, char *argv[])
{
	char *start_mmap, *cur_mmap;
	const char *filename;
	int err, fd, sizeof_int, *reg, *regptr, regfd, regsz, fullfd, emptyfd, what;
	struct stat fbuf;
	unsigned int offset;
	unsigned long start, pages, filesz;

	offset = 0;
	err = 1;
	sizeof_int = sizeof(unsigned int);
	if(argc != 3) {
		printf("usage: %s <pkmem pages> <filename>\n", argv[0]);
		goto out;
	}
	
	pages = atoll(argv[1]);
	filename = argv[2];

	// Open, mmap, obtain size of the file that we will need to send to consumer
	fd = open(filename, O_RDONLY);
	if(fd < 0) { perror("open"); goto out; }
	if(fstat(fd, &fbuf) < 0) { perror("fstat"); goto out; }
	filesz = fbuf.st_size;
	start_mmap = mmap(NULL, filesz, PROT_READ, MAP_PRIVATE, fd, 0);
	if(start_mmap == MAP_FAILED) { perror("mmap"); goto out; }
	cur_mmap = start_mmap;

	// Create the share region and set the status at the first byte
	start = pk_alloc(pages);
	if(start == -1) { pkerror("pk_alloc"); goto out; }
	SET_STATUS(start, EMPTY);
	if(pk_end() < 0) { pkerror("pk_end"); goto out; }
	
	// Create common registry so consumer can know what the start addr is
	regsz = PK_SHM_SIZE;
	reg = create_registry(PK_SHM_NAME, &regfd, &regsz);
	if(reg == NULL) goto out;

	// Write <start>, <pages>, <filesize> into registry
	regptr = reg;
	*((unsigned long *) regptr) = start;
	regptr += sizeof(unsigned long);
	*((unsigned long *) regptr) = pages;
	regptr += sizeof(unsigned long);
	*((unsigned long *) regptr) = filesz;
	regptr += sizeof(unsigned long);

	// Create the FIFOs
        if(umask(0) < 0) { perror("umask"); goto out; }
        if(mknod(FULL_FIFO, S_IFIFO | 0666, 0) < 0) { perror("mknod"); }
        if(mknod(EMPTY_FIFO, S_IFIFO | 0666, 0) < 0) { perror("mknod"); }

	// Open the FIFO
	fullfd = open(FULL_FIFO, O_RDWR);
	if(fullfd < 0) { perror("open"); goto out; }
	emptyfd = open(EMPTY_FIFO, O_RDWR);
	if(emptyfd < 0) { perror("open"); goto out; }

	printf("pkMem share region alloc: %lu, pages %lu\n", start, pages); fflush(stdout);

	// Transfer the data
	// - Wait for consumer to grab what we've put
	// - Signal that we've put stuff
	while(1) {
		if(pk_get(start, PK_FL_WRLCK) < 0) { pkerror("pk_get"); goto out; }
		if(GET_STATUS(start) & FULL) {
			dbg("before WAIT");
			if(pk_end() < 0) { pkerror("pk_end"); goto out; }
			if(read(emptyfd, &what, sizeof(int)) < 0) { perror("read"); goto out; }
			dbg("after WAIT");
			continue;
		}
		if(cur_mmap - start_mmap >= filesz) {
			SET_STATUS(start, COMPLETED);
			dbg("before completed SIGNAL");
			if(pk_end() < 0) { pkerror("pk_end"); goto out; }
			if(write(fullfd, &what, sizeof(int)) < 0) { perror("write"); goto out; }
			dbg("after completed SIGNAL");
			break;
		}

		offset = ((pages * PAGE_SIZE) - sizeof_int);
		if(cur_mmap - start_mmap + offset > filesz)
			offset = filesz - (cur_mmap - start_mmap);

		memcpy((char *)(start + sizeof_int), cur_mmap, offset);
		SET_STATUS(start, FULL);
		SET_SIZE(start, offset);
		dbg("before SIGNAL");
		if(pk_end() < 0) { pkerror("pk_end"); goto out; }
		if(write(fullfd, &what, sizeof(int)) < 0) { perror("write"); goto out; }
		dbg("after SIGNAL");
		cur_mmap += offset;
	}
	// Wait for consumer to have finished
	usleep(300);

	// Remove mmap-ing and close the file
	if(munmap(start_mmap, filesz) < 0) { perror("munmap"); goto out; }
	if(close(fd) < 0) { perror("close"); goto out; }

	// Destroy the registry
	if(destroy_registry(PK_SHM_NAME, reg, &regsz) < 0) goto out;

	// Close fifos
	if(close(emptyfd) < 0) { perror("close"); goto out; }
	if(close(fullfd) < 0) { perror("close"); goto out; }

	// Free the share region
	if(pk_free(start) < 0) { pkerror("pk_free"); goto out; }
	if(pk_end() < 0) { pkerror("pk_end"); goto out; }
	printf("pkMem share region freed: %lu, pages %lu\n", start, pages);

	err = 0;
out:
	return err;
}
